!****************************************************************************
!	NON-GAUSSIAN BAYESIAN GEOSTATISTICAL MODELLING
!	M.B. Palacios and M.F.J. Steel
!	25/6/2004
!****************************************************************************
!
!  PROGRAM: GLGmodel
!
!  Main steps: 
!	1.- Reading parameter file 
!	2.- Reading data set
!	3.- Design matrix
!	4.- Distance matrix
!	5.- Correlation matrix
!	6.- Initial values
!		
!	7.-1st loop: GLG Model
!		7.1.- Drawing beta 
!		7.2.- Drawing sigma^{-2} 
!		7.3.- Drawing omega^2 
!		7.4.- Drawing theta_2 (smoothness parameter)
!		7.5.- Drawing theta_1 (range parameter)
!		7.6.- Drawing epsilon 
!		7.7.- Drawing lambda
!		7.8.- Drawing nu 
!
!	8.-2nd loop: Gaussian Model
!		8.1.- Drawing beta 
!		8.2.- Drawing sigma^{-2} 
!		8.3.- Drawing omega^2 
!		8.4.- Drawing theta_2 (smoothness parameter)
!		8.5.- Drawing theta_1 (range parameter)
!
!	9.- Bayes factors for lambda_i=1 (Formula 13)
!	10.- Moments
!	11.- Marginal likelihood
!	12.- Writing results
!	13.- Writing the file to calculate the semivariogram
!	14.- Prediction
!
!
!	N=100	Maximum number of observations
!   MV=10	Maximum number of variables
!	KH=10	Maximum number of trend parameters
!	MC1=10	Maximum number of clusters
!	GRID=450 Maximum number of points to predict over a grid 
!	npdf=40 Maximum number of points to evaluate the predictive posterior distribution
!
!	P: Number of observations
!	K: Number of trend parameters (INPUT:without covariate OUTPUT: with covariate)
!	K1: Trend degree for the coordinate x
!	K2: Trend degree for the coordinate y
!	simu: number of draws
!	sib: simu-burnin
!	sim: number of saved draws
!	DIST: Distance matrix
!	CORR: Correlation matrix
!	X: Design matrix
!*******************************************************************

PROGRAM GLGmodel

	
	USE NUMERICAL_LIBRARIES

	IMPLICIT NONE

	DOUBLE PRECISION, PARAMETER :: pi=3.1416
	INTEGER, PARAMETER:: N=100, MV=10, KH=10, MC1=10, GRID=450, npdf=40

	LOGICAL :: FILE, ACTION
	INTEGER :: IOS, UNIT, I, J, T, IOSI
	CHARACTER :: STR*30, DATAFILE*30
	DOUBLE PRECISION :: ALLDATA(N,MV)


	INTEGER :: P, K
	INTEGER :: R, sib, simu, burnin, K1,K2,ISEED,R1, sim
	DOUBLE PRECISION :: matern, DIST(N,N), CORR(N,N), X(N,KH)
	DOUBLE PRECISION :: phi, thet1

	DOUBLE PRECISION ::  Z(N), BET(KH), XB(N), RES(N)
	DOUBLE PRECISION :: ICORR(N,N), deter

	DOUBLE PRECISION :: Msig, Vsig, Mbet(KH), Varbet1(KH,KH)
	DOUBLE PRECISION :: gamma1, deter1, Mrang, Vrang
	
	DOUBLE PRECISION :: lnu, LAM(N), Mlambda(N), Vlambda(N,N)
	DOUBLE PRECISION :: Mthet2, Vthet2, logphi1, Mlnu, Vlnu
	
	INTEGER:: NCLUS(MC1), S, MC, MA, MI 
	DOUBLE PRECISION :: omega2,	EPSI(N), lamchange(MC1)
	DOUBLE PRECISION :: Mepsi(N), Vepsi(N,N), RLAM(MC1), REPSI(MC1)

	DOUBLE PRECISION :: XTX(KH,KH),EPSILAM(N),omegacan,REPSILAM(N)
	DOUBLE PRECISION :: tau2, Momega,Vomega, Ms,Vs,Mtau,Vtau
	DOUBLE PRECISION :: thet2,mdis,VDIS(N*(N-1)/2)

	INTEGER :: IJ,NMISS, omgig,nugig,margi
	DOUBLE PRECISION :: XLO,XHI, Mdeter, canlnu,xlamnu
	DOUBLE PRECISION :: gamchange,lnuchange,omegachange,thet2change,thet1change

	DOUBLE PRECISION :: delta, int4,repla,llik, LIK(1500000) 
	DOUBLE PRECISION :: LOGLAM(N), MICORR(N,N)

	DOUBLE PRECISION :: c1, c2, c3, c4, c5, clam45, c45, c6, c7, clam67, c67, c8, c9, c8sav !Hyperparameters
	CHARACTER :: INIV*25, HYP*25, LIKEL*25, VGRAM*25,PREDIC*25,BAYF*25

	DOUBLE PRECISION :: maxi,lag(50), ampli, varg(100000,50),varg1(50)
	
	INTEGER :: np, cvar, vsim, R2, vari, minlag,npr, pred,pl
	DOUBLE PRECISION :: maxdis,XMbet(N), PDIST1(grid,N)	
	DOUBLE PRECISION :: XP(grid,KH) 
	DOUBLE PRECISION :: plag(npdf),ZP1(grid),ZP2(grid),ZP3(grid)
	DOUBLE PRECISION :: fnorm(npdf,grid), FNORM1(npdf,grid),VZP(grid),MZP(grid),VZP1(grid),MZP1(grid),rx(2), ry(2)
	DOUBLE PRECISION :: XRANGE, YRANGE, NGRID, plmin,plmax

	INTEGER :: gaussian
	DOUBLE PRECISION :: TIT,GC2,GC3

	INTEGER :: IB,NC,BF,betap, PRIORTHETA

	DOUBLE PRECISION :: B12,sumb12

	DOUBLE PRECISION :: mz,mz2,vz,mx(KH),mx2(KH),varx(KH),pp

	INTEGER :: CV 

	INTEGER :: BF1, NF, IB1(10), NC1(10),IK(10), IFCOV,COV,Kp,OLStrend

	
	DOUBLE PRECISION :: den,Plam(grid),CPO(grid,N),c88
	DOUBLE PRECISION :: Mrho,Vrho,rho,Msnu,Vsnu,snu,Mtsnu,Vtsnu,tsnu
	DOUBLE PRECISION :: rz(2),oz,sz,PI2

	PI2=1/dsqrt(2.0d0*3.14d0)
	
	OLStrend=0

!******************************************************************
!Reading initial values and hyperparameters from the parameter file
!******************************************************************

!OPEN(UNIT=4,FILE="datafiles/topo.par.txt",STATUS="OLD", ACTION="READ", IOSTAT=ios)
 OPEN(UNIT=4,FILE="datafiles/hot.par.txt",STATUS="OLD", ACTION="READ", IOSTAT=ios)
!OPEN(UNIT=4,FILE="datafiles/cold.par.txt",STATUS="OLD", ACTION="READ", IOSTAT=ios)

	IF(ios /=0) THEN
		PRINT *, "Error during OPENing the data file. Please, try again"
	ELSE 
		PRINT *, "OK"
	END IF

	READ(4,*) DATAFILE
	READ(4,*) P, CV, MC
	READ(4,*) MI, MA
	READ(4,*) (NCLUS(I),I=1,MC)
	READ(4,*) K1,K2,K
	READ(4,*) den
	READ(4,*) IFCOV,COV
	READ(4,*) iseed

	CALL RNSET(ISEED)

	READ(4,*) simu
	READ(4,*) burnin
	READ(4,*) gaussian

	PRINT *, datafile
	PRINT *,'Seed=',iseed
	PRINT *,'Number of simulations=',simu
	PRINT *,'burnin=',burnin

	IF(gaussian==0) PRINT *,'Non-Gaussian case'
	IF(gaussian==1) PRINT *,'Gaussian case'
	 
	READ(4,'(A25)') INIV !Initial values
	PRINT *, INIV

	READ(4,*) gamma1
	READ(4,*) omega2
	READ(4,*) thet2
	READ(4,*) thet1
	READ(4,*) lnu

	PRINT *, gamma1,omega2,thet1,thet2,lnu

	READ(4,'(a25)') HYP !Hyperparameters
	PRINT *, HYP
  
	READ(4,*) c1,betap	!If betap=1 then new prior for betas
	READ(4,*) c2,c3
	READ(4,*) omgig
	READ(4,*) clam45, c4,c5, c45
	READ(4,*) nugig
	READ(4,*) clam67, c6,c7, c67
	READ(4,*) c8
	READ(4,*) c9

!Prior on theta 
	READ(4,*) PRIORTHETA
	
	c8sav=c8

!Marginal Likelihood
	READ(4,'(a25)') LIKEL 
	PRINT *, LIKEL

	READ(4,*) margi, delta

	IF(margi==1) PRINT *, 'delta=', delta

!Bayes factors
	READ(4,'(a25)') BAYF 
	PRINT *, BAYF

	READ(4,*) BF,IB,NC !BF=0 or 1 (1 to calculate Bayes factors) IB=lambda_i=1;NC=number of the cluster

	IF(BF==1) PRINT *, 'Bayes factor at location',IB, 'in cluster',NC

	READ(4,*) BF1, NF !BF1=0 or 1 (1 to save the drawn lambdas) NF=number of obs.s to save their lambdas

	IF(NF.GT.0) THEN
		DO I=1,NF
			READ(4,*) IB1(I), NC1(I)
		END DO
	END IF

!Semivariogram
	READ(4,'(a25)') VGRAM 
	PRINT *, VGRAM

	READ(4,*) cvar, np
	
	IF(cvar==1) PRINT *, 'Number of lags=',np

	READ(4,*)  minlag, maxdis
	
	PRINT *,  minlag, maxdis
	
	READ(4,*) vari

!Prediction
	READ(4,*) PREDIC
	READ(4,*) pred, npr, ngrid
	READ(4,*) pl, plmin,plmax
	IF(npr.gt.0) THEN
		IF(IFCOV.eq.0) THEN
			DO I=1,npr
				READ(4,*) zp1(I), zp2(I)
			END DO
		ELSE
			DO I=1,npr
				READ(4,*) zp1(I), zp2(I), zp3(I)
			END DO
		END IF
	END IF


CLOSE(4)


Kp=K

!********************************
!IK: index for the lambdas to be saved

IF(BF1.eq.1.and.NF.GT.0) THEN
	DO I=1,NF
		IF(NC1(I).EQ.1) THEN
			IK(I)=IB1(I)
		ELSE
			IK(I)=IB1(I)
			DO J=1,NC1(I)-1
				IK(I)=IK(I)+NCLUS(J)
			END DO
		END IF
	END DO
END IF

!***********************************************************
!More Initial values
!***********************************************************
	
DO I=1,P
	LAM(I)= 1.0d0
	EPSI(I)=0.0d0 
END DO

bet=(/8.16,-0.18,-0.44,0.70,0.02,-0.22,0.0,0.0,0.0,0.0/)
phi=1/thet1

!***********************************************************
!Hyperparameters
!***********************************************************

GC2=C2+P/2.0D0		

IF(omgig==0) THEN
	clam45=1.0d0			!lambda
	c4=0.0d0				!delta
	c5=dsqrt(2.0d0*c45)		!gamma
END IF

IF(nugig==0) THEN
	clam67=1.0d0		 !lambda	
	c6=0.0d0			 !delta
	c7=dsqrt(2.0d0*c67)  !gamma
END IF

!***********************************************************
!Reading the data file
!***********************************************************

	CALL READDATA2(DATAFILE,ALLDATA,P,N,MV,STR)

!***********************************************************
!Design matrix
!***********************************************************

	CALL DESIGN1(ALLDATA,P,N,MV,K,KH,K1,K2,X, XTX,rx,ry)

!***********************************************************
!Distance matrix
!***********************************************************

	CALL DISTANCE1(ALLDATA,P,N,MV,DIST)

!***********************************************************
!Median distance
!***********************************************************
	
	IJ=0
	DO I=1,P-1
		DO J=I+1,P
			IJ=IJ+1
			VDIS(IJ)=DIST(I,J)
		END DO 
	END DO


	CALL DEQTIL(IJ,VDIS,1,0.5D0,mdis,XLO,XHI,NMISS)

	PRINT *, ij,  MDIS

c88=c8*dsqrt(2.0d0)/mdis  

!***********************************************************
!Prediction
!***********************************************************

IF(pred.eq.2) THEN
	
	npr=ngrid**2  !Number of points to predict

!Building a grid

	xrange=(rx(2)-rx(1))/(ngrid-1)
	yrange=(ry(2)-ry(1))/(ngrid-1)

	DO I=1,ngrid
		DO J=1,ngrid
			ZP1(J+ngrid*(I-1))=rx(1)+(J-1)*xrange
			ZP2(J+ngrid*(I-1))=rx(1)+(I-1)*yrange
		END DO
	END DO

END IF
 

IF(pred.ne.0) THEN

	CALL prepred1(ALLDATA,P,N,MV,K,KH,K1,K2,npr,ZP1,ZP2,XP,PDIST1)

END IF

IF(pred.eq.1) THEN
	plag(1)=plmin
	DO I=2,pl
		plag(I)=plag(I-1)+(plmax-plmin)/(pl-1)
	END DO
END IF


!***********************************************************
!Scaling data
!***********************************************************


	DO I=1,P
		Z(I)=ALLDATA(I,CV)/den
	END DO

!***********************************************************
!Scaling the covariate
!***********************************************************

IF(IFCOV==1) THEN 
	
	K=K+1

	DO I=1,2
		rz(I)=ALLDATA(1,COV)

	END DO

	DO I=2,P
		IF (ALLDATA(I,COV).LT.rz(1)) rz(1)=ALLDATA(I,COV) !Min COVARIATE
		IF (ALLDATA(I,COV).GT.rz(2)) rz(2)=ALLDATA(I,COV) !Max COVARIATE
	END DO	
	
	oz =(rz(2) + rz(1))/2
	sz = 2/(rz(2) - rz(1))
	
	DO I=1,P
		X(I,K)= sz * (ALLDATA(I,COV) - oz)
	END DO


	!X'X=XTX

	CALL DMXTXF(P,K,X,N,K,XTX,KH)

	IF(pred.eq.1) THEN
		DO I=1,npr
			XP(I,K)= sz * (ZP3(I) - oz)
		END DO
	END IF

END IF

!***********************************************************
!Prior distribution for beta
!***********************************************************

DO I=1,K
	varx(I)=1.0d0
END DO

IF(betap.eq.1) THEN
	pp=1.0d0*P/(P-1)
	
	!mz= sample mean of z
	!vz= sample variance of z

	mz=0.0d0
	mz2=0.0d0
	DO I=1,P
		mz=mz+Z(I)
		mz2=mz2+Z(I)**2
	END DO
	mz=mz/P
	mz2=mz2/P

	vz=pp*(mz2-mz**2)
	
	!varx=sample variance of X

	DO J=1,K
		mx(J)=0.0d0
		mx2(J)=0.0d0
		DO I=1,P
			mx(J)=mx(J)+X(I,J)
			mx2(J)=mx2(J)+X(I,J)**2
		END DO
		mx(J)=mx(J)/P
		mx2(J)=mx2(J)/P
		varx(J)=pp*(mx2(J)-mx(J)**2)
	END DO


	OPEN(unit=15, file='results/varx.txt',status='unknown') 
		WRITE(15,*) 'mz=',mz
		WRITE(15,*) 'vz=',vz
		WRITE(15,*) 'varx='
		WRITE(15,*) (varx(I),I=1,k)
	CLOSE(15)

END IF


!***********************************************************
!Correlation matrix
!***********************************************************

IF(GAUSSIAN==0) THEN
	 
     CALL CORRELATION6(P,N,DIST,thet2,phi,CORR,ICORR,deter)

ELSE
	 CALL GCORRELATION(P,N,DIST,thet2,phi,omega2,CORR,ICORR,deter)

END IF

!***********************************************************
!Initial values for means and variances
!***********************************************************

	Msig=0.00d0
	Vsig=0.00d0
	Mrang=0.00d0
	Vrang=0.00d0
	Mthet2=0.00d0
	Vthet2=0.00d0
	Mlnu=0.00d0
	Vlnu=0.00d0
	Momega=0.00d0
	Vomega=0.00d0
	Ms=0.00d0
	Vs=0.00d0
	Mrho=0.00d0
	Vrho=0.00d0
	Msnu=0.00d0
	Vsnu=0.00d0
	Mtsnu=0.00d0
	Vtsnu=0.00d0
	Mtau=0.00d0
	Vtau=0.00d0


	DO J=1,K
		Mbet(J)=0.00d0
	END DO

	DO J=1,P
		Mlambda(J)=0.00d0
		Mepsi(J)=0.00d0
	END DO


!***********************************************************
!Opening external files to write results
!***********************************************************

IF(BF.EQ.1) BF1=0  !Not to save lambdas 

OPEN(unit=3, file='results/resulmat.txt',status='unknown') 

IF(margi==1) OPEN(unit=7, file='results/margilike.txt',status='unknown') 

OPEN(unit=8, file='results/beta.txt',status='unknown') 

IF(BF1.EQ.1.and.Gaussian==0) OPEN(unit=10, file='results/lambdas.txt',status='unknown') 

IF(BF.EQ.1) OPEN(unit=12, file='results/bayesfactor.txt',status='unknown') 

!***********************************************************
!Initial values
!***********************************************************

R1=0
R2=0

DO  i=1,MC
	lamchange(i)=0
END DO

gamchange=0
thet2change=0
thet1change=0
lnuchange=0
omegachange=0

sib=simu-burnin

B12=0.00d0
sumb12=0.00d0

!******************************************************
!LOOP  gaussian=0   free lambdas
!******************************************************
IF(gaussian==0) THEN

	DO R=1,SIMU

		DO I=1,P
			EPSILAM(I)=EPSI(I)/dsqrt(LAM(I))
		END DO

		tau2=omega2/gamma1  !w2*sigma^2

!******************************************************
!.-Sampling from the conditional of beta
!******************************************************

		CALL HBETA1(P,N,K,KH,EPSILAM,Z,X,XTX,tau2,gamma1,c1,BET,mz,vz,varx,betap)

		CALL DMURRV(P,K,X,N,K,BET,1,P,XB)

		DO I=1,P
			RES(I)=Z(I)-XB(I)
		END DO

!*******************************************************
!.-Sampling from the marginal of sigma^{-2}=gamma1
!*******************************************************

		CALL ISIGMA(P,EPSILAM,omega2,RES,c2,c3,R,burnin,gamchange,gamma1)

!*******************************************************
!.-Sampling from the marginal of omega^2
!*******************************************************

		CALL omegdraw(P,N,RES,EPSILAM,ICORR,gamma1,R,burnin,omegachange,c4,c5,clam45,omega2)

!*******************************************************
!.-Sampling theta_2, the smoothness parameter 
!*******************************************************

		CALL THETA2(P,N,LAM,EPSI,lnu,phi,DIST,CORR,ICORR,deter,thet2,R,burnin,thet2change,c9,PRIORTHETA,C88)

!*******************************************************
!.-Drawing theta_1, the range
!*******************************************************
	
		IF(PRIORTHETA.EQ.1) c8=c8sav*dsqrt(2.0d0*thet2)
	
		CALL THETA1(P,N,LAM,EPSI,lnu,thet2,DIST,CORR,ICORR,deter,c8,mdis,phi,thet1,R,burnin,thet1change)

		rho=2.0d0*thet1*dsqrt(thet2)
	
!*******************************************************
!.-Sampling epsilon
!*******************************************************

		CALL EPSILONS(P,N,CORR,LAM,OMEGA2,RES,gamma1,EPSI)

!*******************************************************
!.-Sampling lambdas
!*******************************************************

		IF(BF.EQ.0) THEN
			CALL LAMBDA(NCLUS, P,N, OMEGA2,CORR,ICORR,EPSI,LNU, gamma1, RES, LAM, lamchange, MC,R, burnin, MA, MI)
		ELSE    
			!Calculating the lambdas for one cluster
			CALL LAMBDA1(IB,NC,NCLUS, P,N, OMEGA2,CORR,ICORR,EPSI,LNU, gamma1, RES, LAM, lamchange, MC,R, burnin, MA, MI, B12)
		END IF

!*******************************************************
!.-Sampling lnu
!*******************************************************

		CALL nudraw(P,N,LAM,ICORR,R,burnin,lnuchange,xlamnu,lnu,c6,c7,clam67)

!*******************************************************
!*******************************************************

		IF(R.GT.burnin) THEN

!*******************************************************
!Marginal likelihood
!*******************************************************

			IF(margi==1) THEN

				CALL MARGCORR(P,N,LAM,CORR,omega2,Mdeter,MICORR)

				repla=DBLINF(P,P,MICORR,N,RES,RES)

				llik=-P*dlog(2*pi)+P*dlog(gamma1)-dlog(Mdeter)-gamma1*repla

				LIK(R-burnin)=dexp(0.5*llik)

				WRITE (7,*) LIK(R-burnin)

			END IF
		
			R1=R1+1

			IF(R1.eq.5) THEN
	
				sim=sim+1

!*************************************************************
!Writing results in external files
!*************************************************************

				WRITE (3,'(7F20.15)') gamma1, omega2, thet2, thet1, lnu, rho, dsqrt(tau2)

				WRITE (8,'(7F15.7)') (BET(I),I=1,k)

				IF(BF1.eq.1.and.NF.GT.0)  WRITE (10,'(10F20.15)') (LAM(IK(I)),I=1,NF)

!*************************************************************
! Calculation of moments - Sums
!*************************************************************

!E*[isigma]=Msig, Var*[isigma]=Vsig, E*[B]=Mbet
!E*[range]=Mrang and Var*[range]=Vrang
!E*[thet2]=Mthet2 and Var*[thet2]=Vthet2

				snu=1/gamma1*dexp(lnu)
				tsnu=dsqrt(omega2)/dexp(lnu*0.5d0)

				Msig=Msig+gamma1
				Vsig=Vsig+gamma1**2
				Mrang=Mrang+thet1
				Vrang=Vrang+thet1**2
				Mthet2=Mthet2+thet2
				Vthet2=Vthet2+thet2**2
				Mlnu=Mlnu+lnu
				Vlnu=Vlnu+lnu**2
				Momega=Momega+omega2
				Vomega=Vomega+omega2**2
				Ms=Ms+1/dsqrt(gamma1)
				Vs=Vs+1/gamma1
				Mrho=Mrho+rho
				Vrho=Vrho+rho**2
				Msnu=Msnu+snu
				Vsnu=Vsnu+snu**2
				Mtsnu=Mtsnu+tsnu
				Vtsnu=Vtsnu+tsnu**2
				Mtau=Mtau+dsqrt(tau2)
				Vtau=Vtau+tau2

				DO J=1,K
					Mbet(J)=Mbet(J)+BET(J)
				END DO

				DO J=1,P
					Mlambda(J)=Mlambda(J)+LAM(J)
					Mepsi(J)=Mepsi(J)+EPSI(J)
				END DO

!Covariance matrix of Beta
!E[BB']

				DO i=1,k
					DO j=1,k
						Varbet1(i,j)=Varbet1(i,j)+BET(i)*BET(j)
					END DO	
				END DO

!Variance matrix of Lambda

				DO i=1,P
					DO j=1,P
						Vlambda(i,j)=Vlambda(i,j)+LAM(i)*LAM(j)
						Vepsi(i,j)=Vepsi(i,j)+EPSI(i)*EPSI(j)
					END DO
				END DO

!Bayes factors

				sumb12=sumb12+B12

!*************************************************************
!Prediction
!*************************************************************

				IF(pred.ne.0) THEN

					CALL RNGET(ISEED)

					!p(z_p|z_o,rest)

					CALL predicgrid(P,N,npr,PDIST1,thet2,phi,ICORR,LAM,lnu, &
						K, KH, CORR, omega2,gamma1, X,XP,bet,Z, MZP,VZP)

					IF(pred.eq.1) THEN

					!Evaluating the normal distribution

					DO I=1,pl
						DO J=1,npr
							FNorm(I,J)=PI2*dsqrt(1/VZP(J))*dexp(-(PLAG(I)-MZP(J))**2/(VZP(J)*2.0D0))
						END DO
					END DO

					DO I=1,pl
						DO J=1,npr
							Fnorm1(I,J)=Fnorm1(I,J)+Fnorm(I,J)
						END DO
					END DO

				ELSE IF(pred.eq.2) THEN

					DO I=1,npr
						MZP1(I)=MZP1(I)+MZP(I)
						VZP1(I)=VZP1(I)+dsqrt(VZP(I))
					END DO

				END IF
	
				CALL RNSET(ISEED)

			END IF

!*************************************************************

			R1=0

		END IF	!R1

!*************************************************************
!Semivariogram 
!*************************************************************

		IF(cvar==1) THEN
			R2=R2+1
			IF(R2.eq.20) THEN
				vsim=vsim+1
				R2=0
			END IF
		END IF

!*******************************

		END IF	!R  > burnin
	
	END DO

!*******************************************************
! END LOOP gaussian=0 free lambdas
!*******************************************************

ELSE

!*******************************************************
! LOOP gaussian=1 Gaussian case
!*******************************************************

	DO R=1,SIMU
	
!******************************************************
!.-Sampling from the conditional of beta
!******************************************************

		CALL GHBETA1(P,N,K,KH,Z,X,ICORR,gamma1,c1,BET,mz,vz,varx,betap)

		CALL DMURRV(P,K,X,N,K,BET,1,P,XB)

		DO I=1,P
			RES(I)=Z(I)-XB(I) !t=Z-XB
		END DO

		
!*******************************************************
!.-Sampling from the marginal of sigma^{-2}=gamma1
!*******************************************************
	
!t'*ISIGMA*t=(Z-XB)'(C+w2*I_k)^{-1}(Z-XB)

		TIT=DBLINF(P,P,ICORR,N,RES,RES)
	
		GC3=C3+TIT/2.0D0

		CALL DRNGAM(1,GC2,GAMMA1)
	
		GAMMA1=GAMMA1/GC3

!*******************************************************
!.-Sampling from the marginal of omega^2
!*******************************************************

		CALL Gomegdraw(P,N,DIST,thet2,phi,RES,gamma1,R,burnin,omegachange,c4,c5,clam45,omega2,CORR,ICORR,deter)

!*******************************************************
!.-Sampling theta_2, the smoothness parameter 
!*******************************************************

		CALL GTHETA2(P,N,RES,phi,DIST,CORR,ICORR,deter,thet2,R,burnin,thet2change,gamma1,omega2,c9,PRIORTHETA,c88)

!*******************************************************
!.-Drawing theta_1, the range
!*******************************************************

		IF(PRIORTHETA.EQ.1) c8=c8sav*dsqrt(2.0d0*thet2)
	
		CALL GTHETA1(P,N,RES,thet2,DIST,CORR,ICORR,deter,c8,mdis,phi,thet1,R,burnin,thet1change,omega2,gamma1)


		rho=2.0d0*thet1*dsqrt(thet2)

!*******************************************************
!*******************************************************

		IF(R.GT.burnin) THEN

!*******************************************************
!Marginal likelihood
!*******************************************************

			IF(margi==1) THEN

				repla=DBLINF(P,P,ICORR,N,RES,RES)

				llik=-P*dlog(2*pi)+P*dlog(gamma1)-dlog(deter)-gamma1*repla

				LIK(R-burnin)=dexp(0.5*llik)

				WRITE (7,*) LIK(R-burnin)

			END IF
		
			R1=R1+1

			IF(R1.eq.5) THEN
	
				sim=sim+1

!*******************************************************
!Writing results in external files
!*******************************************************

				tau2=omega2/gamma1  !w2*sigma^2

				WRITE (3,'(6F20.15)') gamma1, omega2, thet2, thet1, rho, dsqrt(tau2)

  				WRITE(8,'(7F15.7)') (BET(I),I=1,k)

!*******************************************************
! Calculation of moments - Sums
!*******************************************************

!E*[isigma]=Msig, Var*[isigma]=Vsig, E*[B]=Mbet
!E*[range]=Mrang and Var*[range]=Vrang
!E*[thet2]=Mthet2 and Var*[thet2]=Vthet2

				snu=1/gamma1
				tsnu=dsqrt(omega2)

				Msig=Msig+gamma1
				Vsig=Vsig+gamma1**2
				Mrang=Mrang+thet1
				Vrang=Vrang+thet1**2
				Mthet2=Mthet2+thet2
				Vthet2=Vthet2+thet2**2
				Momega=Momega+omega2
				Vomega=Vomega+omega2**2
				Ms=Ms+1/dsqrt(gamma1)
				Vs=Vs+1/gamma1
				Mrho=Mrho+rho
				Vrho=Vrho+rho**2
				Msnu=Msnu+snu
				Vsnu=Vsnu+snu**2
				Mtsnu=Mtsnu+tsnu
				Vtsnu=Vtsnu+tsnu**2
				Mtau=Mtau+dsqrt(tau2)
				Vtau=Vtau+tau2

				DO J=1,K
					Mbet(J)=Mbet(J)+BET(J)
				END DO


!Covariance matrix of Beta

!E[BB']

				DO i=1,k
					DO j=1,k
						Varbet1(i,j)=Varbet1(i,j)+BET(i)*BET(j)
					END DO	
				END DO

!*************************************************************
!Prediction
!*************************************************************

				IF(pred.ne.0) THEN

					CALL RNGET(ISEED)

					!p(z_p|z_o,rest)
	
					CALL gpredicgrid(P,N,npr,PDIST1,thet2,phi,ICORR, &
						K, KH, CORR, omega2,gamma1,X,XP,bet,Z,MZP,VZP)

					IF(pred.eq.1) THEN

						!Evaluating the normal distribution

						DO I=1,pl
							DO J=1,npr
								FNorm(I,J)=PI2*dsqrt(1/VZP(J))*dexp(-(PLAG(I)-MZP(J))**2/(VZP(J)*2.0D0))
							END DO
						END DO

						DO I=1,pl
							DO J=1,npr
								Fnorm1(I,J)=Fnorm1(I,J)+Fnorm(I,J)
							END DO
						END DO

					ELSE IF(pred.eq.2) THEN

						DO I=1,npr
							MZP1(I)=MZP1(I)+MZP(I)
							VZP1(I)=VZP1(I)+dsqrt(VZP(I))
						END DO

					END IF

					CALL RNSET(ISEED)

				END IF

!*************************************************************

				R1=0

			END IF	!R1

!*************************************************************
!Semivariogram
!*************************************************************

			IF(cvar==1) THEN
				R2=R2+1
				IF(R2.eq.20) THEN
					vsim=vsim+1
					R2=0
				END IF
			END IF

!*******************************

		END IF	!R  > burnin
	
	END DO

!*******************************************************
!END LOOP gaussian =1  Gaussian case
!*******************************************************

END IF


!********************************************************
!Closing external files
!********************************************************

 CLOSE(3)  
 
 IF(margi==1) CLOSE(7)      

 CLOSE(8)      

IF(gaussian==0) THEN
	
	DO i=1,MC
		lamchange(i)=lamchange(i)/sib
	END DO

	gamchange=gamchange/sib
	lnuchange=lnuchange/sib
END IF

thet2change=thet2change/sib
thet1change=thet1change/sib
omegachange=omegachange/sib

IF(gaussian==0) THEN
	PRINT *, 'Proportion of acceptance for gamma1=', gamchange
	PRINT *, 'Proportion of acceptance for nu=', lnuchange
END IF

PRINT *, 'Proportion of acceptance for theta2=', thet2change
PRINT *, 'Proportion of acceptance for theta1=', thet1change
PRINT *, 'Proportion of acceptance for omega2=', omegachange


!Bayes factors

	sumb12=sumb12/sim

IF(BF.EQ.1) THEN
	WRITE(12,*) 'Expectancy for',IB,'in cluster',NC,'=',sumb12
END IF

!*******************************************************
!Moments
!*************************************************************

IF(gaussian==0) THEN

	CALL MOMENT6(P,N,K,KH,sim,Msig,Vsig,Mbet,Varbet1,Mrang,Vrang, &
		Mlambda, Vlambda, Mthet2, Vthet2, Mlnu,Vlnu,   &
		Mepsi, Vepsi, Momega,Vomega,Ms,Vs,Mrho,Vrho,Msnu,Vsnu,Mtsnu,Vtsnu,Mtau,Vtau)

ELSE

	!Moments

	!E*[isigma]=Msig, Var*[isigma]=Vsig, E*[B]=Mbet
	!E*[range]=Mrang and Var*[range]=Vrang


	Msig=Msig/sim
	Vsig=Vsig/sim-Msig**2
	Mrang=Mrang/sim
	Vrang=Vrang/sim-Mrang**2
	Mthet2=Mthet2/sim
	Vthet2=Vthet2/sim-Mthet2**2
	Momega=Momega/sim
	Vomega=Vomega/sim-Momega**2
	Ms=Ms/sim
	Vs=Vs/sim-Ms**2
	Mrho=Mrho/sim
	Vrho=Vrho/sim-Mrho**2
	Msnu=Msnu/sim
	Vsnu=Vsnu/sim-Msnu**2
	Mtsnu=Mtsnu/sim
	Vtsnu=Vtsnu/sim-Mtsnu**2
	Mtau=Mtau/sim
	Vtau=Vtau/sim-Mtau**2

	DO i=1,k
		Mbet(i)=Mbet(i)/sim
	END DO

	!Covariance matrix of Beta
	!E[BB']

	DO i=1,k
		DO j=1,k
			Varbet1(i,j)=Varbet1(i,j)/sim-Mbet(i)*Mbet(j)
		END DO
	END DO

END IF

!*************************************************************
!Marginal likelihood p(x|Model)
!*************************************************************

IF(margi==1) THEN

	CALL MARGLIKE(LIK,sib,delta,int4)
	
END IF

!*************************************************************
! Writing results
!*************************************************************

IF(gaussian==0) THEN

	CALL   WRITING5(P, N, K, KH, simu, burnin, sim, Msig,Vsig,Mbet,Varbet1,Mrang, Vrang, &
			ISEED,Mlambda,Vlambda, STR, Mthet2, Vthet2,&
			Mlnu, Vlnu, MC, lamchange, Mepsi, Vepsi, Momega,Vomega, Ms,Vs, &
			gamchange,thet2change,thet1change,c1,c2,c3,clam45,c4,c5,c45,clam67,c6,c7,c67,c8sav,c9,lnuchange,omegachange,int4, &
			omgig,nugig, priortheta,Mrho,Vrho,Msnu,Vsnu,Mtsnu,Vtsnu,IFCOV,Mtau,Vtau)
ELSE

	CALL  WRITING6(P, N, K, KH, simu, burnin, sim, Msig,Vsig,Mbet,Varbet1,Mrang, Vrang, &
			ISEED, STR, Mthet2, Vthet2, Momega,Vomega, Ms,Vs, &
			thet2change,thet1change,c1,c2,c3,clam45,c4,c5,c45,clam67,c6,c7,c67,c8sav,c9,omegachange,int4, &
			omgig, priortheta,Mrho,Vrho,Msnu,Vsnu,Mtsnu,Vtsnu,IFCOV,Mtau,Vtau)

END IF

!*************************************************************
!Semivariogram
!*************************************************************

IF(cvar==1) THEN

OPEN(unit=11, file='results/outfile.txt',status='unknown')

	WRITE(11,*) DATAFILE
	WRITE(11,*) 'datafiles\beta.txt    '
	WRITE(11,*) 'datafiles\OLSbeta.txt    '
	WRITE(11,*) 'datafiles\resulmat.txt'


	WRITE(11,14) P,CV, '\P=N obs,CV=vble column'
14	FORMAT(1X, I2,1X,I2,8X, A40)

	WRITE(11,16) K1,K2,Kp,'\k1=trend x, k2=trend y, K=N parameters'
16	FORMAT(1X, I2,1X,I2,1X,I2,8X, A40)

	WRITE(11,19) DEN,'\DEN=scale data'
19	FORMAT(1X, F10.3, 8X, A40)


	WRITE(11,18) IFCOV,COV, '\IFCOV=1 covariate, COVA=Covariate column'
18	FORMAT(1X, I2,1X,I2,8X,A40)

 	WRITE(11,15) GAUSSIAN, '\0=Non-Gaussian 1=Gaussian'

	WRITE(11,15) vsim,'\Number of saved betas'
	WRITE(11,15) sim,'\Number of saved drawings'
	WRITE(11,15) np,'\Number of points for the semivariogram'
	WRITE(11,17) maxdis,'\From 0 to 1: Maximum distance'
17	FORMAT(1X, F4.2, 8X, A40)

	WRITE(11,15) minlag,'\Minimum number of pairs per lag'
	WRITE(11,15) vari,'\0=Matheron 1=Cressie&Hawkins'
 	WRITE(11,15) OLStrend, '\OLStrend=1 ols, 0 median'

15	FORMAT(1X, I10, 8X, A40)

CLOSE(11)

END IF

!*************************************************************
!Prediction
!*************************************************************

IF(pred.eq.1) THEN

	OPEN(unit=12, file='results/ppd.txt',status='unknown') 

	DO I=1,pl
		DO J=1,npr
			Fnorm1(I,J)=Fnorm1(I,J)/sim
		END DO
		WRITE(12,'(100F15.10)') plag(I),(Fnorm1(I,J),J=1,npr)
	END DO

	CLOSE(12)

ELSE IF(pred.eq.2) THEN

	OPEN(unit=13, file='results/ppdmeansd.txt',status='unknown') 

	WRITE(13,'(4A10)') 'X','Y','MEAN','SD'

	DO I=1,npr
		MZP1(I)=MZP1(I)/sim
		VZP1(I)=VZP1(I)/sim
		WRITE(13,'(4F10.4)')  ZP1(I), ZP2(I), MZP1(I),VZP1(I)
	END DO

CLOSE(13)

END IF

END PROGRAM GLGmodel